/*
 * Copyright 2018 ScaleDynamics SAS. All rights reserved.
 * Licensed under the MIT license.
 */

declare module "warp" {

    /**
     * Start a parallel execution and provides its result as a Promise to be resolved.   
     *   
     * Perform a call equivalent to `fnct(...args)` in a
     * [parallel flow](https://github.com/ScaleDynamics/warp/blob/master/README.md#parallel-flow).
     *   
     * The [Execution context](https://github.com/ScaleDynamics/warp/blob/master/README.md#execution-context)
     * sent to the parallel flow is captured at the time of the call to `warp.callAsPromise`.
     *   
     * The [Parallel flow](https://github.com/ScaleDynamics/warp/blob/master/README.md#parallel-flow)
     * ends when:   
     * - `fnct` has returned to caller   
     * - and, if its returned value was a `Promise`, it is [fulfilled or rejected](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise#Description).   
     * This triggers a capture of its
     * [execution context](https://github.com/ScaleDynamics/warp/blob/master/README.md#execution-context).
     * Then, in a regular asynchronous task of the main flow, changes made by the parallel execution
     * are applied and the `Promise` returned by `callAsPromise` is rejected or resolved with the
     * result of the parallel flow.
     *    
     * If something prevents execution in a
     * [parallel flow](https://github.com/ScaleDynamics/warp/blob/master/README.md#parallel-flow),
     * then the returned `Promise` is rejected with an error.
     * 
     * If the parallel task throws an error, then the Promise returned by `callAsPromise` is
     * rejected with this error as reason.
     *
     * @returns a `Promise` that will be resolved or rejected (according to `fnct` result) when
     * `fnct` parallel execution is complete.
     * @param fnct Function to execute in parallel.
     * @param args A list of arguments to be passed to the function.
     *    
     * See [warp README](https://github.com/ScaleDynamics/warp/blob/master/README.md) for more
     * details.
     */
    function callAsPromise(fnct: (...args: any[]) => any, ...args: any[]): Promise;

    /**
     * Start a parallel execution of a function with callback.   
     *   
     * Perform a call equivalent to `fnct(...args)` in a
     * [parallel flow](https://github.com/ScaleDynamics/warp/blob/master/README.md#parallel-flow).
     * Last element of `args` must be a callback.
     *   
     * The [Execution context](https://github.com/ScaleDynamics/warp/blob/master/README.md#execution-context)
     * sent to the parallel flow is captured at the time of the call to `warp.call`.
     *   
     * The [Parallel flow](https://github.com/ScaleDynamics/warp/blob/master/README.md#parallel-flow)
     * ends when `fnct` has returned to caller and the callback was called.
     * This triggers a capture of its
     * [execution context](https://github.com/ScaleDynamics/warp/blob/master/README.md#execution-context).
     * Then, in the main flow, changes made by the parallel execution are applied and the callback
     * is called.   
     *    
     * If something prevents execution in a
     * [parallel flow](https://github.com/ScaleDynamics/warp/blob/master/README.md#parallel-flow),
     * then `call` will throw an Error.
     * 
     * If the parallel task throws an error, then the error is thrown back in a regular asynchronous
     * context of the main flow. 
     *
     * @param fnct Function to execute in parallel.
     * @param args A list of arguments to be passed to the function. The last of those arguments
     * must be a `callback`.
     * @param callback The callback passed to the function. It must be given as last element of
     * `args`. This `callback` is not executed in the
     * parallel context, but in the main javascript flow like a regular asynchronous task.   
     *    
     * See [StarNode help page](https://github.com/ScaleDynamics/warp/blob/master/README.md) for more
     * details.
     */
    function call(fnct: (...args: any[]) => any, ...args: any[]): void;

    /**
     * Provide the number of _Warp_ threads currently instanciate by the _Warp_ runtime.
     *
     * @return The number of _Warp_ threads.
     */
    function getCurrentNumWarpThreads(): number;
    
    /**
     * Provide the list of _Warp_ features supported by the current _Warp_ runtime.
     *
     * @return A structured object.
     */
    function getRuntimeCapabilities(): Object;
}
